# Generated by Django 2.2.2 on 2019-07-17 04:14

import uuid

import _socket
import django.contrib.auth.models
import django.core.validators
import django.db.models.deletion
import django.db.models.manager
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django.utils.timezone
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models

import edc_action_item.models.action_item
import edc_registration.models.managers
import edc_sites.models
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("edc_registration", "0017_auto_20190305_0123"),
        ("edc_lab", "0019_auto_20190305_0123"),
        ("edc_action_item", "0022_auto_20190628_2136"),
        ("auth", "0011_update_proxy_permissions"),
        ("edc_visit_schedule", "0006_auto_20190706_0706"),
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="DataDictionary",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("model", models.CharField(max_length=250)),
                ("model_verbose_name", models.CharField(max_length=250, null=True)),
                ("number", models.IntegerField()),
                ("prompt", models.TextField()),
                ("field_name", models.CharField(max_length=250)),
                ("field_type", models.CharField(max_length=250, null=True)),
                ("max_length", models.IntegerField(null=True)),
                ("decimal_places", models.IntegerField(null=True)),
                ("max_digits", models.IntegerField(null=True)),
                ("nullable", models.BooleanField(null=True)),
                ("default", models.CharField(max_length=250, null=True)),
                ("help_text", models.CharField(max_length=250, null=True)),
                ("active", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name": "Data Dictionary Item",
                "verbose_name_plural": "Data Dictionary Items",
                "ordering": ("model", "number", "prompt"),
                "default_permissions": ("view",),
                "unique_together": {("model", "field_name")},
            },
        ),
        migrations.CreateModel(
            name="DataManagerActionItem",
            fields=[],
            options={
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_action_item.actionitem",),
            managers=[
                ("objects", django.db.models.manager.Manager()),
                ("on_site", edc_action_item.models.action_item.CurrentSiteManager()),
            ],
        ),
        migrations.CreateModel(
            name="DataManagerUser",
            fields=[],
            options={
                "verbose_name": "Data Manager User",
                "verbose_name_plural": "Data Manager Users",
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("auth.user",),
            managers=[("objects", django.contrib.auth.models.UserManager())],
        ),
        migrations.CreateModel(
            name="QuerySubject",
            fields=[],
            options={
                "verbose_name": "Subject",
                "verbose_name_plural": "Subjects",
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_registration.registeredsubject",),
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                (
                    "objects",
                    edc_registration.models.managers.RegisteredSubjectManager(),
                ),
            ],
        ),
        migrations.CreateModel(
            name="QueryUser",
            fields=[],
            options={
                "verbose_name": "Data Query User",
                "verbose_name_plural": "Data Query Users",
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("auth.user",),
            managers=[("objects", django.contrib.auth.models.UserManager())],
        ),
        migrations.CreateModel(
            name="QueryVisitSchedule",
            fields=[],
            options={
                "verbose_name": "Visit Schedule",
                "verbose_name_plural": "Visit Schedule",
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_visit_schedule.visitschedule",),
        ),
        migrations.CreateModel(
            name="RequisitionPanel",
            fields=[],
            options={
                "verbose_name": "Requisition Panel",
                "verbose_name_plural": "Requisition Panels",
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_lab.panel",),
        ),
        migrations.CreateModel(
            name="HistoricalQueryRule",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("title", models.CharField(db_index=True, max_length=150)),
                (
                    "query_text",
                    models.TextField(
                        blank=True,
                        help_text="Generic query text for auto-generated queries.",
                        null=True,
                    ),
                ),
                (
                    "reference_model",
                    models.CharField(editable=False, max_length=150, null=True),
                ),
                (
                    "reference_date",
                    models.CharField(
                        choices=[
                            ("report_datetime", "Report datetime (visit report)"),
                            ("drawn_datetime", "Specimen draw date (requisition)"),
                        ],
                        default="report_datetime",
                        editable=False,
                        max_length=25,
                    ),
                ),
                ("timing", models.IntegerField(default=48, verbose_name="Timing")),
                (
                    "timing_units",
                    models.CharField(
                        choices=[
                            ("hours", "Hours"),
                            ("days", "Days"),
                            ("weeks", "Weeks"),
                            ("months", "Months"),
                        ],
                        default="hours",
                        max_length=25,
                        verbose_name="Units",
                    ),
                ),
                (
                    "query_priority",
                    models.CharField(
                        choices=[("high", "High"), ("NORMAL", "Normal")],
                        default="NORMAL",
                        max_length=25,
                        verbose_name="Priority",
                    ),
                ),
                (
                    "rule_handler_name",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("default", "Default"),
                            (
                                "lumbar_puncture_q13",
                                "Lumbar Puncture (Q13, 15, 21, 23, 24)",
                            ),
                        ],
                        default="default",
                        max_length=150,
                        null=True,
                    ),
                ),
                (
                    "reference",
                    models.CharField(db_index=True, default=uuid.uuid4, max_length=36),
                ),
                ("comment", models.TextField(blank=True, null=True)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "requisition_panel",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Requisition will be expected on day of visit.",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_data_manager.RequisitionPanel",
                        verbose_name="Are responses linked to a requisition? If so, which",
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_data_manager.DataManagerUser",
                        verbose_name="Query raised by",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Query Rule",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalDataQuery",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("action_identifier", models.CharField(db_index=True, max_length=50)),
                (
                    "parent_action_identifier",
                    models.CharField(
                        help_text="action identifier that links to parent reference model instance.",
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "related_action_identifier",
                    models.CharField(
                        help_text="action identifier that links to related reference model instance.",
                        max_length=30,
                        null=True,
                    ),
                ),
                ("action_item_reason", models.TextField(editable=False, null=True)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "report_datetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="Query date"
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(editable=False, max_length=50, null=True),
                ),
                ("title", models.CharField(blank=True, max_length=150, null=True)),
                (
                    "query_priority",
                    models.CharField(
                        choices=[("high", "High"), ("NORMAL", "Normal")],
                        default="NORMAL",
                        max_length=25,
                        verbose_name="Priority",
                    ),
                ),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        default=0,
                        help_text="Defaults to '0'. For example, when combined with the visit code `1000` would make `1000.0`.",
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(25),
                        ],
                        verbose_name="Visit code sequence",
                    ),
                ),
                (
                    "timepoint",
                    models.DecimalField(decimal_places=1, max_digits=6, null=True),
                ),
                (
                    "query_text",
                    models.TextField(help_text="Describe the query in detail."),
                ),
                (
                    "site_resolved_datetime",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="Site resolved on"
                    ),
                ),
                ("site_response_text", models.TextField(blank=True, null=True)),
                (
                    "site_response_status",
                    models.CharField(
                        choices=[
                            ("New", "New"),
                            ("open", "Open"),
                            ("feedback", "Feedback"),
                            ("resolved", "Resolved"),
                        ],
                        default="New",
                        max_length=25,
                        verbose_name="Site status",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("open", "Open"),
                            ("resolved", "Resolved"),
                            ("resolved_with_action", "Resolved, with plan of action"),
                        ],
                        default="open",
                        max_length=25,
                        verbose_name="TCC status",
                    ),
                ),
                (
                    "resolved_datetime",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="TCC resolved on"
                    ),
                ),
                (
                    "plan_of_action",
                    models.TextField(
                        blank=True,
                        help_text="If required, provide a plan of action",
                        null=True,
                    ),
                ),
                (
                    "locked",
                    models.BooleanField(
                        default=False,
                        help_text="If locked, this query will NEVER be reopened.",
                    ),
                ),
                (
                    "rule_generated",
                    models.BooleanField(
                        default=False,
                        help_text="This query was auto-generated by a query rule.",
                    ),
                ),
                (
                    "rule_reference",
                    models.CharField(
                        default=uuid.uuid4,
                        max_length=150,
                        null=True,
                        verbose_name="Query rule reference",
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "parent_action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "registered_subject",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_data_manager.QuerySubject",
                        verbose_name="Subject Identifier",
                    ),
                ),
                (
                    "related_action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "requisition_panel",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Requisition will be expected on day of visit.",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_data_manager.RequisitionPanel",
                        verbose_name="Are responses linked to a requisition? If so, which",
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Please select a name from the list",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_data_manager.DataManagerUser",
                        verbose_name="Query raised by",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sites.Site",
                    ),
                ),
                (
                    "tcc_user",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Please select a name from the list",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_data_manager.DataManagerUser",
                        verbose_name="TCC resolved by",
                    ),
                ),
                (
                    "visit_schedule",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_data_manager.QueryVisitSchedule",
                        verbose_name="Visit",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Data Query",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalDataDictionary",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("model", models.CharField(max_length=250)),
                ("model_verbose_name", models.CharField(max_length=250, null=True)),
                ("number", models.IntegerField()),
                ("prompt", models.TextField()),
                ("field_name", models.CharField(max_length=250)),
                ("field_type", models.CharField(max_length=250, null=True)),
                ("max_length", models.IntegerField(null=True)),
                ("decimal_places", models.IntegerField(null=True)),
                ("max_digits", models.IntegerField(null=True)),
                ("nullable", models.BooleanField(null=True)),
                ("default", models.CharField(max_length=250, null=True)),
                ("help_text", models.CharField(max_length=250, null=True)),
                ("active", models.BooleanField(default=False)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Data Dictionary Item",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="DataQuery",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("action_identifier", models.CharField(max_length=50, unique=True)),
                (
                    "parent_action_identifier",
                    models.CharField(
                        help_text="action identifier that links to parent reference model instance.",
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "related_action_identifier",
                    models.CharField(
                        help_text="action identifier that links to related reference model instance.",
                        max_length=30,
                        null=True,
                    ),
                ),
                ("action_item_reason", models.TextField(editable=False, null=True)),
                (
                    "report_datetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="Query date"
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(editable=False, max_length=50, null=True),
                ),
                ("title", models.CharField(blank=True, max_length=150, null=True)),
                (
                    "query_priority",
                    models.CharField(
                        choices=[("high", "High"), ("NORMAL", "Normal")],
                        default="NORMAL",
                        max_length=25,
                        verbose_name="Priority",
                    ),
                ),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        default=0,
                        help_text="Defaults to '0'. For example, when combined with the visit code `1000` would make `1000.0`.",
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(25),
                        ],
                        verbose_name="Visit code sequence",
                    ),
                ),
                (
                    "timepoint",
                    models.DecimalField(decimal_places=1, max_digits=6, null=True),
                ),
                (
                    "query_text",
                    models.TextField(help_text="Describe the query in detail."),
                ),
                (
                    "site_resolved_datetime",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="Site resolved on"
                    ),
                ),
                ("site_response_text", models.TextField(blank=True, null=True)),
                (
                    "site_response_status",
                    models.CharField(
                        choices=[
                            ("New", "New"),
                            ("open", "Open"),
                            ("feedback", "Feedback"),
                            ("resolved", "Resolved"),
                        ],
                        default="New",
                        max_length=25,
                        verbose_name="Site status",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("open", "Open"),
                            ("resolved", "Resolved"),
                            ("resolved_with_action", "Resolved, with plan of action"),
                        ],
                        default="open",
                        max_length=25,
                        verbose_name="TCC status",
                    ),
                ),
                (
                    "resolved_datetime",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="TCC resolved on"
                    ),
                ),
                (
                    "plan_of_action",
                    models.TextField(
                        blank=True,
                        help_text="If required, provide a plan of action",
                        null=True,
                    ),
                ),
                (
                    "locked",
                    models.BooleanField(
                        default=False,
                        help_text="If locked, this query will NEVER be reopened.",
                    ),
                ),
                (
                    "rule_generated",
                    models.BooleanField(
                        default=False,
                        help_text="This query was auto-generated by a query rule.",
                    ),
                ),
                (
                    "rule_reference",
                    models.CharField(
                        default=uuid.uuid4,
                        max_length=150,
                        null=True,
                        verbose_name="Query rule reference",
                    ),
                ),
                (
                    "action_item",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "data_dictionaries",
                    models.ManyToManyField(
                        blank=True,
                        help_text="select all that apply",
                        to="edc_data_manager.DataDictionary",
                        verbose_name="CRF question(s)",
                    ),
                ),
                (
                    "parent_action_item",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "recipients",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Select any additional recipients. Users in the `Site Data Manager` group are automatically included.",
                        related_name="_dataquery_recipients_+",
                        to="edc_data_manager.QueryUser",
                        verbose_name="Sent to",
                    ),
                ),
                (
                    "registered_subject",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_data_manager.QuerySubject",
                        verbose_name="Subject Identifier",
                    ),
                ),
                (
                    "related_action_item",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "requisition_panel",
                    models.ForeignKey(
                        blank=True,
                        help_text="Requisition will be expected on day of visit.",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="edc_data_manager.RequisitionPanel",
                        verbose_name="Are responses linked to a requisition? If so, which",
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        help_text="Please select a name from the list",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="edc_data_manager.DataManagerUser",
                        verbose_name="Query raised by",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="sites.Site",
                    ),
                ),
                (
                    "tcc_user",
                    models.ForeignKey(
                        blank=True,
                        help_text="Please select a name from the list",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="tcc_user",
                        to="edc_data_manager.DataManagerUser",
                        verbose_name="TCC resolved by",
                    ),
                ),
                (
                    "visit_schedule",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_data_manager.QueryVisitSchedule",
                        verbose_name="Visit",
                    ),
                ),
            ],
            options={
                "verbose_name": "Data Query",
                "verbose_name_plural": "Data Queries",
            },
            managers=[("on_site", edc_sites.models.CurrentSiteManager())],
        ),
        migrations.CreateModel(
            name="CrfDataDictionary",
            fields=[],
            options={
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_data_manager.datadictionary",),
        ),
        migrations.CreateModel(
            name="RequisitionDataDictionary",
            fields=[],
            options={
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_data_manager.datadictionary",),
        ),
        migrations.CreateModel(
            name="VisitDataDictionary",
            fields=[],
            options={
                "proxy": True,
                "default_permissions": ("view",),
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_data_manager.datadictionary",),
        ),
        migrations.CreateModel(
            name="QueryRule",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("title", models.CharField(max_length=150, unique=True)),
                (
                    "query_text",
                    models.TextField(
                        blank=True,
                        help_text="Generic query text for auto-generated queries.",
                        null=True,
                    ),
                ),
                (
                    "reference_model",
                    models.CharField(editable=False, max_length=150, null=True),
                ),
                (
                    "reference_date",
                    models.CharField(
                        choices=[
                            ("report_datetime", "Report datetime (visit report)"),
                            ("drawn_datetime", "Specimen draw date (requisition)"),
                        ],
                        default="report_datetime",
                        editable=False,
                        max_length=25,
                    ),
                ),
                ("timing", models.IntegerField(default=48, verbose_name="Timing")),
                (
                    "timing_units",
                    models.CharField(
                        choices=[
                            ("hours", "Hours"),
                            ("days", "Days"),
                            ("weeks", "Weeks"),
                            ("months", "Months"),
                        ],
                        default="hours",
                        max_length=25,
                        verbose_name="Units",
                    ),
                ),
                (
                    "query_priority",
                    models.CharField(
                        choices=[("high", "High"), ("NORMAL", "Normal")],
                        default="NORMAL",
                        max_length=25,
                        verbose_name="Priority",
                    ),
                ),
                (
                    "rule_handler_name",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("default", "Default"),
                            (
                                "lumbar_puncture_q13",
                                "Lumbar Puncture (Q13, 15, 21, 23, 24)",
                            ),
                        ],
                        default="default",
                        max_length=150,
                        null=True,
                    ),
                ),
                (
                    "reference",
                    models.CharField(default=uuid.uuid4, max_length=36, unique=True),
                ),
                ("comment", models.TextField(blank=True, null=True)),
                (
                    "recipients",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Select any additional recipients. Users in the `Site Data Manager` group are automatically included.",
                        related_name="_queryrule_recipients_+",
                        to="edc_data_manager.QueryUser",
                        verbose_name="Send to",
                    ),
                ),
                (
                    "requisition_panel",
                    models.ForeignKey(
                        blank=True,
                        help_text="Requisition will be expected on day of visit.",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="edc_data_manager.RequisitionPanel",
                        verbose_name="Are responses linked to a requisition? If so, which",
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="edc_data_manager.DataManagerUser",
                        verbose_name="Query raised by",
                    ),
                ),
                (
                    "sites",
                    models.ManyToManyField(
                        help_text="Leave blank to apply to all.", to="sites.Site"
                    ),
                ),
                (
                    "visit_schedule",
                    models.ManyToManyField(
                        blank=True,
                        help_text="select all that apply",
                        related_name="_queryrule_visit_schedule_+",
                        to="edc_data_manager.QueryVisitSchedule",
                        verbose_name="Visit",
                    ),
                ),
                (
                    "visit_schedule_exclude",
                    models.ManyToManyField(
                        blank=True,
                        help_text="select all that apply",
                        related_name="_queryrule_visit_schedule_exclude_+",
                        to="edc_data_manager.QueryVisitSchedule",
                        verbose_name="Visit (exclude)",
                    ),
                ),
                (
                    "data_dictionaries",
                    models.ManyToManyField(
                        blank=True,
                        help_text="select all that apply",
                        related_name="_queryrule_data_dictionaries_+",
                        to="edc_data_manager.CrfDataDictionary",
                        verbose_name="Question(s)",
                    ),
                ),
            ],
            options={
                "verbose_name": "Query Rule",
                "verbose_name_plural": "Query Rules",
                "ordering": ("title",),
            },
        ),
        migrations.AddIndex(
            model_name="dataquery",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "action_identifier",
                    "title",
                    "rule_reference",
                    "registered_subject",
                    "visit_schedule",
                ],
                name="edc_data_ma_subject_875fa1_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="queryrule",
            index=models.Index(
                fields=["title", "active"], name="edc_data_ma_title_c3118f_idx"
            ),
        ),
    ]



from .appointment_creator import AppointmentCreator
from .appointments_creator import AppointmentsCreator
from .unscheduled_appointment_creator import UnscheduledAppointmentCreator
from .utils import create_next_appointment_as_interim, create_unscheduled_appointment

__all__ = [
    "AppointmentCreator",
    "UnscheduledAppointmentCreator",
    "create_next_appointment_as_interim",
    "create_unscheduled_appointment",
]



# Generated by Django 5.2.6 on 2025-09-17 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_locator", "0039_alter_historicalsubjectlocator_revision_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="action_identifier",
            field=models.CharField(
                blank=True, db_index=True, default="", max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="consent_model",
            field=models.CharField(blank=True, default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="consent_version",
            field=models.CharField(blank=True, default="", max_length=10),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="may_sms",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                default="",
                max_length=25,
                verbose_name="Has the participant given permission <b>to be contacted by SMS</b> by study staff for follow-up purposes during the study?",
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="action_identifier",
            field=models.CharField(blank=True, default="", max_length=50, unique=True),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="action_item_reason",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="consent_model",
            field=models.CharField(blank=True, default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="consent_version",
            field=models.CharField(blank=True, default="", max_length=10),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="may_sms",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                default="",
                max_length=25,
                verbose_name="Has the participant given permission <b>to be contacted by SMS</b> by study staff for follow-up purposes during the study?",
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                default="",
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
            ),
        ),
    ]



from uuid import UUID

from clinicedc_constants import UUID_PATTERN, YES
from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from edc_visit_tracking.utils import get_related_visit_model_cls

from edc_lab.admin.fieldsets import (
    requisition_identifier_fields,
    requisition_verify_fields,
)


class RequisitionAdminMixin:
    default_item_type = "tube"
    default_item_count = 1
    default_estimated_volume = 5.0

    ordering = ("requisition_identifier",)

    date_hierarchy = "requisition_datetime"

    radio_fields = {  # noqa: RUF012
        "is_drawn": admin.VERTICAL,
        "reason_not_drawn": admin.VERTICAL,
        "item_type": admin.VERTICAL,
    }

    search_fields: tuple[str, ...] = (
        "requisition_identifier",
        "subject_identifier",
        "panel__display_name",
    )

    @staticmethod
    def visit_code(obj=None) -> str:
        return f"{obj.related_visit.visit_code}.{obj.related_visit.visit_code_sequence}"

    @staticmethod
    def requisition(obj=None):
        if obj.is_drawn == YES:
            return obj.requisition_identifier
        if not obj.is_drawn:
            return format_html(
                '<span style="color:red;">{}</span>', obj.requisition_identifier
            )
        return format_html(
            "{html}",
            html=mark_safe('<span style="color:red;">not drawn</span>'),  # nosec B703, B308
        )

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "panel":
            panel_id = request.GET.get("panel")
            if panel_id and UUID_PATTERN.match(panel_id):
                kwargs["queryset"] = db_field.related_model.objects.filter(
                    pk=UUID(request.GET.get("panel"))
                )
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def get_list_filter(self, request) -> tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = ("requisition_datetime", "site", "is_drawn", "panel")
        list_filter = tuple(f for f in list_filter if f not in custom_fields)
        return custom_fields + list_filter

    def get_list_display(self, request) -> tuple[str, ...]:
        list_display = super().get_list_display(request)
        custom_fields = (
            "requisition",
            "subject_identifier",
            "visit_code",
            "panel",
            "requisition_datetime",
            "hostname_created",
        )
        list_display = tuple(f for f in list_display if f not in custom_fields)
        return custom_fields + list_display

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return tuple(
            set(readonly_fields + requisition_identifier_fields + requisition_verify_fields)
        )

    def get_changeform_initial_data(self, request) -> dict:
        initial_data = super().get_changeform_initial_data(request)
        if isinstance(request.GET.get(self.model.related_visit_model_attr()), UUID):
            try:
                related_visit = get_related_visit_model_cls().objects.get(
                    id=request.GET.get(self.model.related_visit_model_attr())
                )
            except ObjectDoesNotExist:
                # TODO: how do we get here? PRN?
                pass
            else:
                initial_data.update(
                    requisition_datetime=related_visit.report_datetime,
                    item_type=self.default_item_type,
                    item_count=self.default_item_count,
                    estimated_volume=self.default_estimated_volume,
                )
        return initial_data

        # if self.fields.get("specimen_type"):
        #     self.fields["specimen_type"].widget.attrs["readonly"] = True



# Generated by Django 2.1 on 2018-10-22 19:04

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_notification", "0002_update_notifications"),
        ("edc_auth", "0005_userprofile_job_title"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="notifications",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to={"enabled": True},
                to="edc_notification.Notification",
            ),
        )
    ]



from django.urls.conf import path

from edc_dashboard.url_names import url_names
from edc_protocol.research_protocol_config import ResearchProtocolConfig

from .admin_site import edc_adverse_event_admin
from .views import (
    AeHomeView,
    ClosedTmgAeListboardView,
    NewTmgAeListboardView,
    OpenTmgAeListboardView,
    TmgHomeView,
)
from .views import DeathListboardView as TmgDeathListboardView
from .views import SummaryListboardView as TmgSummaryListboardView

app_name = "edc_adverse_event"


urlpatterns = NewTmgAeListboardView.urls(
    namespace=app_name,
    url_names_key="new_tmg_ae_listboard_url",
    identifier_label="subject_identifier",
    identifier_pattern=ResearchProtocolConfig().subject_identifier_pattern,
)
urlpatterns += OpenTmgAeListboardView.urls(
    namespace=app_name,
    url_names_key="open_tmg_ae_listboard_url",
    identifier_label="subject_identifier",
    identifier_pattern=ResearchProtocolConfig().subject_identifier_pattern,
)
urlpatterns += ClosedTmgAeListboardView.urls(
    namespace=app_name,
    url_names_key="closed_tmg_ae_listboard_url",
    identifier_label="subject_identifier",
    identifier_pattern=ResearchProtocolConfig().subject_identifier_pattern,
)
urlpatterns += TmgDeathListboardView.urls(
    namespace=app_name,
    url_names_key="tmg_death_listboard_url",
    identifier_label="subject_identifier",
    identifier_pattern=ResearchProtocolConfig().subject_identifier_pattern,
)
urlpatterns += TmgSummaryListboardView.urls(
    namespace=app_name,
    url_names_key="tmg_summary_listboard_url",
    identifier_label="subject_identifier",
    identifier_pattern=ResearchProtocolConfig().subject_identifier_pattern,
)
urlpatterns += [
    path("tmg/", TmgHomeView.as_view(), name="tmg_home_url"),
    path("ae/", AeHomeView.as_view(), name="ae_home_url"),
    path("admin/", edc_adverse_event_admin.urls),
    path("", AeHomeView.as_view(), name="home_url"),
]

url_names.register(key="tmg_home_url", url_with_namespace=f"{app_name}:tmg_home_url")
url_names.register(key="ae_home_url", url_with_namespace=f"{app_name}:ae_home_url")



from clinicedc_constants import CLOSED, OPEN, OTHER
from django.db import models
from django.utils import timezone

from edc_sites.model_mixins import SiteModelMixin

from ...constants import STORAGE, TESTING
from ...identifiers import ManifestIdentifier

STATUS = ((OPEN, "Open"), (CLOSED, "Closed"))

MANIFEST_CATEGORY = ((TESTING, "Testing"), (STORAGE, "Storage"), (OTHER, "Other"))


class ManifestModelMixin(SiteModelMixin, models.Model):
    manifest_identifier = models.CharField(
        verbose_name="Manifest Identifier", max_length=25, editable=False, unique=True
    )

    manifest_datetime = models.DateTimeField(default=timezone.now)

    export_datetime = models.DateTimeField(blank=True, null=True)

    export_references = models.TextField(blank=True, default="")

    description = models.TextField(
        verbose_name="Description of contents",
        default="",
        help_text="If blank will be automatically generated",
    )

    status = models.CharField(max_length=15, default=OPEN, choices=STATUS)

    category = models.CharField(max_length=25, default=TESTING, choices=MANIFEST_CATEGORY)

    category_other = models.CharField(max_length=25, default="", blank=True)

    comment = models.TextField(verbose_name="Comment", blank=True, default="")

    shipped = models.BooleanField(default=False)

    printed = models.BooleanField(default=False)

    printed_datetime = models.DateTimeField(null=True, blank=True)

    def save(self, *args, **kwargs):
        if not self.manifest_identifier:
            identifier = ManifestIdentifier()
            self.manifest_identifier = identifier.identifier
        if self.shipped and not self.export_datetime:
            self.export_datetime = timezone.now()
        elif not self.shipped:
            self.export_datetime = None
            self.printed = False
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.manifest_identifier,)

    natural_key.dependencies = ("sites.Site",)

    @property
    def human_readable_identifier(self):
        x = self.manifest_identifier
        return f"{x[0:4]}-{x[4:8]}-{x[8:12]}"

    class Meta:
        abstract = True



from ..constants import (
    ACCOUNT_MANAGER,
    ACCOUNT_MANAGER_ROLE,
    ADMINISTRATION,
    AUDITOR,
    AUDITOR_ROLE,
    CLINIC,
    CLINIC_SUPER,
    CLINICIAN_ROLE,
    CLINICIAN_SUPER_ROLE,
    CUSTOM_ROLE,
    EVERYONE,
    NURSE_ROLE,
    PII,
    PII_VIEW,
    STAFF_ROLE,
    STATISTICIAN,
)


def get_default_roles() -> dict[str, list[str]]:
    """Format {ROLE_NAME: [GROUP_NAME, GROUP_NAME, ...]"""
    return {
        ACCOUNT_MANAGER_ROLE: [ACCOUNT_MANAGER],
        AUDITOR_ROLE: [
            AUDITOR,
            PII_VIEW,
        ],
        CLINICIAN_ROLE: [
            CLINIC,
            PII,
        ],
        CLINICIAN_SUPER_ROLE: [
            CLINIC_SUPER,
            CLINIC,
            PII,
        ],
        CUSTOM_ROLE: [],
        NURSE_ROLE: [
            CLINIC,
            PII,
        ],
        STAFF_ROLE: [ADMINISTRATION, EVERYONE],
        STATISTICIAN: [AUDITOR],
    }



from datetime import date
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.db import models

from ..stubs import SubjectScreeningModelStub


class ScreeningMethodsModeMixin(models.Model):
    def __str__(self: SubjectScreeningModelStub):
        return f"{self.screening_identifier} {self.gender} {self.age_in_years}"

    def natural_key(self: SubjectScreeningModelStub):
        return (self.screening_identifier,)

    @staticmethod
    def get_search_slug_fields():
        return "screening_identifier", "subject_identifier", "reference"

    @property
    def estimated_dob(self: SubjectScreeningModelStub) -> date:
        return self.report_datetime.astimezone(ZoneInfo("UTC")).date() - relativedelta(
            years=self.age_in_years
        )

    class Meta:
        abstract = True



from django.db import models
from django.utils import timezone
from django.utils.translation import gettext as _
from django_crypto_fields.fields import EncryptedTextField

from edc_constants.choices import YES_NO_NA
from edc_identifier.model_mixins import UniqueSubjectIdentifierModelMixin

from .. import site_consents
from ..consent_definition_extension import ConsentDefinitionExtension
from ..exceptions import ConsentExtensionDefinitionModelError

___all__ = ["ConsentExtensionModelMixin"]


class ConsentExtensionModelMixin(UniqueSubjectIdentifierModelMixin, models.Model):
    # declare with an FK your subject consent!
    subject_consent = models.ForeignKey("mysubjectconsent", on_delete=models.PROTECT)

    report_datetime = models.DateTimeField(default=timezone.now)

    agrees_to_extension = models.CharField(
        verbose_name=_(
            "Does the participant give consent to extend clinic "
            "followup as per the protocol amendment?"
        ),
        max_length=15,
        choices=YES_NO_NA,
        help_text=_("See above for the definition of extended followup."),
    )

    comment = EncryptedTextField(verbose_name="Comment", max_length=250, blank=True, null=True)

    consent_extension_version = models.CharField(
        verbose_name="Consent extension version",
        max_length=10,
        default="",
        editable=False,
    )

    consent_extension_definition_name = models.CharField(
        verbose_name="Consent extension definition",
        max_length=50,
        default="",
        editable=False,
    )

    def __str__(self) -> str:
        if self.subject_identifier:
            return f"{self._meta.verbose_name} for {self.subject_identifier}"
        return ""

    def save(self, *args, **kwargs):
        self.subject_identifier = self.subject_consent.subject_identifier
        cdefext = self.consent_definition_extension
        self.consent_extension_version = cdefext.version
        self.consent_extension_definition_name = cdefext.name
        super().save(*args, **kwargs)

    @property
    def consent_definition_extension(self) -> ConsentDefinitionExtension:
        cdef = site_consents.get_consent_definition(
            model=self.subject_consent._meta.label_lower,
            version=self.subject_consent.version,
            report_datetime=self.report_datetime,
        )
        if self._meta.label_lower != cdef.extended_by.model:
            raise ConsentExtensionDefinitionModelError(
                f"Incorrect model for consent definition extension. "
                f"This model cannot be used "
                f"to 'extend' consent version '{cdef.version}'. Expected "
                f"'{cdef.extended_by.model}'. Got '{self._meta.label_lower}'."
            )
        return cdef.extended_by

    class Meta:
        abstract = True



from __future__ import annotations

from collections import Counter
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any

from dateutil.relativedelta import relativedelta
from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from edc_utils import floor_secs, formatted_datetime
from edc_utils.date import to_local

from .baseline import Baseline
from .exceptions import OffScheduleError, OnScheduleError, SiteVisitScheduleError
from .site_visit_schedules import site_visit_schedules

if TYPE_CHECKING:
    from django.db import models
    from edc_appointment.models import Appointment

    from .model_mixins import OnScheduleModelMixin
    from .schedule import Schedule
    from .visit import CrfCollection, Visit
    from .visit_schedule import VisitSchedule


def get_default_max_visit_window_gap():
    return getattr(settings, "EDC_VISIT_SCHEDULE_DEFAULT_MAX_VISIT_GAP_ALLOWED", 7)


def get_enforce_window_period_enabled() -> bool:
    return getattr(settings, "EDC_VISIT_SCHEDULE_ENFORCE_WINDOW_PERIOD", True)


def get_lower_datetime(instance: Appointment) -> datetime:
    """Returns the datetime of the lower window"""
    if instance.related_visit:
        dte = instance.appt_datetime
    else:
        instance.visit.dates.base = instance.first.timepoint_datetime
        dte = instance.visit.dates.lower
    return dte


def get_upper_datetime(instance) -> datetime:
    """Returns the datetime of the upper window"""
    instance.visit.dates.base = instance.first.timepoint_datetime
    return instance.visit.dates.upper


def is_baseline(
    instance: Any = None,
    timepoint: Decimal | None = None,
    visit_code_sequence: int | None = None,
    visit_schedule_name: str | None = None,
    schedule_name: str | None = None,
) -> bool:
    return Baseline(
        instance=instance,
        timepoint=timepoint,
        visit_code_sequence=visit_code_sequence,
        visit_schedule_name=visit_schedule_name,
        schedule_name=schedule_name,
    ).value


def raise_if_baseline(subject_visit) -> None:
    if subject_visit and is_baseline(instance=subject_visit):
        raise forms.ValidationError("This form is not available for completion at baseline.")


def raise_if_not_baseline(subject_visit) -> None:
    if subject_visit and not is_baseline(instance=subject_visit):
        raise forms.ValidationError("This form is only available for completion at baseline.")


def get_onschedule_models(subject_identifier: str, report_datetime: datetime) -> list[str]:
    """Returns a list of onschedule models, in label_lower format,
    for this subject and date.
    """
    onschedule_models = []
    subject_schedule_history_model_cls = django_apps.get_model(
        "edc_visit_schedule.SubjectScheduleHistory"
    )
    for onschedule_model_obj in subject_schedule_history_model_cls.objects.onschedules(
        subject_identifier=subject_identifier, report_datetime=report_datetime
    ):
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            onschedule_model=onschedule_model_obj._meta.label_lower
        )
        onschedule_models.append(schedule.onschedule_model)
    return onschedule_models


def get_offschedule_models(subject_identifier: str, report_datetime: datetime) -> list[str]:
    """Returns a list of offschedule models, in label_lower format,
    for this subject and date.

    Subject status must be ON_SCHEDULE.

    See also, manager method `onschedules`.
    """
    offschedule_models = []
    subject_schedule_history_model_cls = django_apps.get_model(
        "edc_visit_schedule.SubjectScheduleHistory"
    )
    onschedule_models = subject_schedule_history_model_cls.objects.onschedules(
        subject_identifier=subject_identifier, report_datetime=report_datetime
    )
    for onschedule_model_obj in onschedule_models:
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            onschedule_model=onschedule_model_obj._meta.label_lower
        )
        offschedule_models.append(schedule.offschedule_model)
    return offschedule_models


def off_schedule_or_raise(
    subject_identifier=None,
    report_datetime=None,
    visit_schedule_name=None,
    schedule_name=None,
) -> None:
    """Returns True if subject is on the given schedule
    on this date.
    """
    visit_schedule = site_visit_schedules.get_visit_schedule(
        visit_schedule_name=visit_schedule_name
    )
    schedule = visit_schedule.schedules.get(schedule_name)
    if schedule.is_onschedule(subject_identifier, report_datetime):
        raise OnScheduleError(
            f"Not allowed. Subject {subject_identifier} is on schedule "
            f"{visit_schedule.verbose_name}.{schedule_name} on "
            f"{formatted_datetime(report_datetime)}. "
            f"See model '{schedule.offschedule_model_cls().verbose_name}'"
        )


def off_all_schedules_or_raise(subject_identifier: str):
    """Raises an exception if subject is still on any schedule."""
    for visit_schedule in site_visit_schedules.get_visit_schedules().values():
        for schedule in visit_schedule.schedules.values():
            try:
                with transaction.atomic():
                    schedule.onschedule_model_cls.objects.get(
                        subject_identifier=subject_identifier
                    )
            except ObjectDoesNotExist:
                pass
            else:
                try:
                    with transaction.atomic():
                        schedule.offschedule_model_cls.objects.get(
                            subject_identifier=subject_identifier
                        )
                except ObjectDoesNotExist as e:
                    model_name = schedule.offschedule_model_cls()._meta.verbose_name.title()
                    raise OffScheduleError(
                        f"Subject cannot be taken off study. Subject is still on a "
                        f"schedule. Got schedule '{visit_schedule.name}."
                        f"{schedule.name}. "
                        f"Complete the offschedule form `{model_name}` first. "
                        f"Subject identifier='{subject_identifier}', "
                    ) from e
    return True


def offstudy_datetime_after_all_offschedule_datetimes(
    subject_identifier: str,
    offstudy_datetime: datetime,
    exception_cls=None,
) -> None:
    exception_cls = exception_cls or forms.ValidationError
    for visit_schedule in site_visit_schedules.get_visit_schedules().values():
        for schedule in visit_schedule.schedules.values():
            try:
                schedule.onschedule_model_cls.objects.get(
                    subject_identifier=subject_identifier
                )
            except ObjectDoesNotExist:
                pass
            else:
                try:
                    offschedule_obj = schedule.offschedule_model_cls.objects.get(
                        subject_identifier=subject_identifier,
                        offschedule_datetime__gt=offstudy_datetime,
                    )
                except ObjectDoesNotExist:
                    pass
                else:
                    offschedule_datetime = formatted_datetime(
                        offschedule_obj.offschedule_datetime
                    )
                    raise exception_cls(
                        "`Offstudy` datetime cannot be before any "
                        "`offschedule` datetime. "
                        f"Got {subject_identifier} went off schedule "
                        f"`{visit_schedule.name}.{schedule.name}` on "
                        f"{offschedule_datetime}."
                    )


def report_datetime_within_onschedule_offschedule_datetimes(
    subject_identifier: str,
    report_datetime: datetime,
    visit_schedule_name: str,
    schedule_name: str,
    exception_cls=None,
):
    exception_cls = exception_cls or forms.ValidationError
    visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name)
    schedule = visit_schedule.schedules.get(schedule_name)
    try:
        onschedule_obj = schedule.onschedule_model_cls.objects.get(
            subject_identifier=subject_identifier
        )
    except ObjectDoesNotExist as e:
        raise OnScheduleError(
            f"Subject is not on schedule. {visit_schedule_name}.{schedule_name}. "
            f"Got {subject_identifier}"
        ) from e
    try:
        offschedule_obj = schedule.offschedule_model_cls.objects.get(
            subject_identifier=subject_identifier,
            offschedule_datetime__lte=report_datetime,
        )
    except ObjectDoesNotExist:
        offschedule_obj = None
        offschedule_datetime = report_datetime
    else:
        offschedule_datetime = offschedule_obj.offschedule_datetime
    if not (
        floor_secs(onschedule_obj.onschedule_datetime)
        <= floor_secs(report_datetime)
        <= floor_secs(offschedule_datetime)
    ):
        onschedule_datetime = formatted_datetime(onschedule_obj.onschedule_datetime)
        if offschedule_obj:
            offschedule_datetime = formatted_datetime(offschedule_obj.offschedule_datetime)
            error_msg = (
                "Invalid report datetime. Expected a datetime between "
                f"{onschedule_datetime} and {offschedule_datetime}. "
                "See onschedule and offschedule."
            )
        else:
            error_msg = (
                "Invalid report datetime. Expected a datetime on or after "
                f"{onschedule_datetime}. See onschedule."
            )
        raise exception_cls(error_msg)


def get_onschedule_model_instance(
    subject_identifier: str,
    reference_datetime: datetime,
    visit_schedule_name: str,
    schedule_name: str,
) -> OnScheduleModelMixin:
    """Returns the onschedule model instance

    Increment reference_datetime by 1 sec to avoid millisecond
    in lte comparison.
    """
    schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name).schedules.get(
        schedule_name
    )
    model_cls = django_apps.get_model(schedule.onschedule_model)
    try:
        onschedule_obj = model_cls.objects.get(
            subject_identifier=subject_identifier,
            onschedule_datetime__lte=reference_datetime + relativedelta(seconds=1),
        )
    except ObjectDoesNotExist as e:
        dte_as_str = formatted_datetime(to_local(reference_datetime))
        raise OffScheduleError(
            "Subject is not on a schedule. Using subject_identifier="
            f"`{subject_identifier}` and appt_datetime=`{dte_as_str}`. Got {e}"
        ) from e
    return onschedule_obj


def get_duplicates(list_items: list[Any]) -> list:
    return [n for n, count in Counter(list_items).items() if count > 1]


def get_models_from_collection(collection: CrfCollection) -> list[models.Model]:
    return [f.model_cls for f in collection]


def get_proxy_models_from_collection(collection: CrfCollection) -> list[models.Model]:
    return [f.model_cls for f in collection if f.model_cls._meta.proxy]


def get_proxy_root_model(proxy_model: models.Model) -> models.Model | None:
    """Returns proxy's root (concrete) model if `proxy_model` is a
    proxy model, else returns None.
    """
    if proxy_model._meta.proxy:
        return proxy_model._meta.concrete_model
    return None


def check_models_in_visit_schedule() -> dict[str, list]:
    """Try to look up all models IN visit schedule collections
    or add to the errors list.

    used by system_checks.
    """
    if not site_visit_schedules.loaded:
        raise SiteVisitScheduleError("Registry is not loaded.")
    errors = {"visit_schedules": [], "schedules": [], "visits": []}
    for visit_schedule in site_visit_schedules.visit_schedules.values():
        errors["visit_schedules"].extend(check_visit_schedule_models(visit_schedule))
        for schedule in visit_schedule.schedules.values():
            errors["schedules"].extend(check_schedule_models(schedule))
            for visit in schedule.visits.values():
                errors["visits"].extend(check_visit_models(visit))
    return errors


def check_visit_schedule_models(visit_schedule: VisitSchedule) -> list[str]:
    """Try to look up the models declared ON the VisitSchedule
    or add to the errors list.

    Used by system_checks.
    """
    errors = []
    for model in ["death_report", "locator", "offstudy"]:
        try:
            getattr(visit_schedule, f"{model}_model_cls")
        except LookupError as e:
            errors.append(f"{e} See visit schedule '{visit_schedule.name}'.")
    return errors


def check_schedule_models(schedule: Schedule) -> list[str]:
    """Try to look up the models declared ON the Schedule
    or add to errors list.

    Used by system_checks.
    """
    errors = []
    for model in ["onschedule", "offschedule", "appointment"]:
        try:
            getattr(schedule, f"{model}_model_cls")
        except LookupError as e:
            errors.append(f"{e} See visit schedule '{schedule.name}'.")
    return errors


def check_visit_models(visit: Visit):
    """Try to look up all models declared in the Visit
    collections or add to errors list.

    Used by system_checks.
    """
    errors = []
    visit_models = list(set([f.model for f in visit.all_crfs]))
    for model in visit_models:
        try:
            django_apps.get_model(model)
        except LookupError as e:
            errors.append(f"{e} Got Visit {visit.code} crf.model={model}.")
    visit_models = list(set([f.model for f in visit.all_requisitions]))
    for model in visit_models:
        try:
            django_apps.get_model(model)
        except LookupError as e:
            errors.append(f"{e} Got Visit {visit.code} requisition.model={model}.")
    return errors


def allow_unscheduled(appointment: Appointment | None = None):
    """Validate if an unschedule appointment is allowed.

    Defined on the Visit object of the visit schedule.

    Normal case: Visit object  allow_unscheduled=True and there is a
    next appointment and the next appointment date is at least one
    day before the main .0 appointment.

    Special case: The special case would allow an unscheduled
    appointment to follow the last appointment in the schedule. This
    condition is rare and not recommended or at least should be
    considered carefully. If Visit `allow_unscheduled_extended` is
    True and there are no further appointments in the schedule, then
    the unscheduled appointment is allowed.

    See also visit.rupper_extended.
    """
    return (
        not appointment.relative_next and appointment.visit.allow_unscheduled_extended
    ) or (
        appointment.appt_datetime.date() + relativedelta(days=1)
        != appointment.relative_next.appt_datetime.date()
    )



from datetime import datetime
from pathlib import Path

import pandas as pd


class Data:

    def __init__(
        self,
        label: str,
        table_df: pd.DataFrame,
        data_df: pd.DataFrame,
        filename_prefix: str,
        folder: str | None = None,
    ):
        self.label = label
        self.table_df = table_df
        self.data_df = data_df
        self.filename_prefix = filename_prefix
        self.folder = folder or "~/"

    def __repr__(self):
        return f"Data({self.label}) <obs={len(self.data_df)}>"

    def to_csv(
        self, folder: str | None = None, filename: str | None = None, cols: int | None = None
    ):
        folder = folder or self.folder
        cols = cols or 5
        datestamp = datetime.now().strftime("%Y%m%d%H%M")
        filename = filename or f"{self.filename_prefix}_table_{self.label}_{datestamp}.csv"
        path = Path(folder) / filename
        self.table_df.iloc[:, :cols].to_csv(
            path_or_buf=path, encoding="utf-8", index=0, sep="|"
        )



from edc_auth.constants import (
    AUDITOR_ROLE,
    CLINICIAN_ROLE,
    CLINICIAN_SUPER_ROLE,
    NURSE_ROLE,
    TMG_ROLE,
)
from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

REVIEW = "REVIEW"


def update_site_auths():
    site_auths.add_post_update_func(
        "edc_review_dashboard", remove_default_model_permissions_from_edc_permissions
    )

    site_auths.add_custom_permissions_tuples(
        model="edc_review_dashboard.edcpermissions",
        codename_tuples=[
            (
                "edc_review_dashboard.view_subject_review_listboard",
                "Can access subject review listboard",
            )
        ],
    )
    site_auths.add_group("edc_review_dashboard.view_subject_review_listboard", name=REVIEW)

    site_auths.update_role(REVIEW, name=AUDITOR_ROLE)
    site_auths.update_role(REVIEW, name=TMG_ROLE)
    site_auths.update_role(REVIEW, name=CLINICIAN_ROLE)
    site_auths.update_role(REVIEW, name=CLINICIAN_SUPER_ROLE)
    site_auths.update_role(REVIEW, name=NURSE_ROLE)


update_site_auths()



class ActivePrescriptionRefillOverlap(Exception):  # noqa: N818
    pass


class RefillError(Exception):
    pass


class ActiveRefillAlreadyExists(Exception):  # noqa: N818
    pass


class NextRefillError(Exception):
    pass


class NextStudyMedicationError(Exception):
    pass


class InsufficientQuantityError(Exception):
    pass


class PackagingSidMismatchError(Exception):
    pass


class PackagingSubjectIdentifierMismatchError(Exception):
    pass


class PackagingLotNumberMismatchError(Exception):
    pass


class PrescriptionAlreadyExists(Exception):  # noqa: N818
    pass


class PrescriptionError(Exception):
    pass


class PrescriptionExpired(Exception):  # noqa: N818
    pass


class RefillAlreadyExists(Exception):  # noqa: N818
    pass


class PrescriptionNotStarted(Exception):  # noqa: N818
    pass


class RefillCreatorError(Exception):
    pass


class RefillEndDatetimeError(Exception):
    pass


class StudyMedicationError(Exception):
    pass


class StockError(Exception):
    pass


class InsufficientStockError(Exception):
    pass


class InvalidContainer(Exception):  # noqa: N818
    pass


class ReceiveError(Exception):
    pass


class RepackError(Exception):
    pass


class RepackRequestError(Exception):
    pass


class ProcessStockRequestError(Exception):
    pass


class StockRequestError(Exception):
    pass


class ChecksumError(Exception):
    pass


class AllocationError(Exception):
    pass


class OrderItemError(Exception):
    pass


class ReceiveItemError(Exception):
    pass


class StockRequestItemError(Exception):
    pass


class LotError(Exception):
    pass


class StockTransferError(Exception):
    pass


class AssignmentError(Exception):
    pass


class StockTransferConfirmationError(Exception):
    pass


class StorageBinError(Exception):
    pass


class ConfirmAtLocationError(Exception):
    pass



from .get_next_scheduled_visit_for_subjects_df import (
    get_next_scheduled_visit_for_subjects_df,
)
from .in_stock_for_subjects_df import in_stock_for_subjects_df
from .no_stock_for_subjects_df import no_stock_for_subjects_df
from .stock_for_subjects import stock_for_subjects_df



from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import Tags, register
from django.core.management.color import color_style

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_auth"
    verbose_name = "Edc Authentication"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self):
        from edc_appconfig.system_checks import check_for_edc_appconfig  # noqa: PLC0415

        from .system_checks import (  # noqa: PLC0415
            check_auth_updater,
            check_etc_dir,
            check_site_auths,
            check_static_root,
        )

        register(check_etc_dir, Tags.security, deploy=True)
        register(check_static_root, Tags.security, deploy=True)
        register(check_site_auths)
        register(check_auth_updater)
        register(check_for_edc_appconfig)



from django import forms
from django.core.exceptions import ObjectDoesNotExist

from edc_model.utils import model_exists_or_raise

from ..utils import get_clinical_review_baseline_model_cls


class ClinicalReviewBaselineRequiredModelFormMixin:
    """Asserts Baseline Clinical Review exists or raise"""

    def clean(self) -> dict:
        cleaned_data = super().clean()
        model_cls = get_clinical_review_baseline_model_cls()
        if self._meta.model != model_cls and cleaned_data.get("subject_visit"):
            try:
                model_exists_or_raise(
                    subject_visit=cleaned_data.get("subject_visit"),
                    model_cls=model_cls,
                    singleton=True,
                )
            except ObjectDoesNotExist:
                raise forms.ValidationError(
                    f"Complete the `{model_cls._meta.verbose_name}` CRF first."
                )
        return cleaned_data



from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import register

from .system_checks import middleware_check


class AppConfig(DjangoAppConfig):
    name = "edc_dashboard"

    listboard_template_name = None
    include_in_administration_section = False

    def ready(self):
        register(middleware_check)



from clinicedc_constants import INCOMPLETE
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from edc_constants.choices import DOCUMENT_STATUS

from ..update_crf_status_for_instance import update_crf_status_for_instance


def get_crf_status_default():
    return getattr(settings, "CRF_STATUS_DEFAULT", INCOMPLETE)


class CrfStatusModelMixin(models.Model):
    """A model mixin that adds CRF status fields and
    a method to update the status.

    The method is called in an edc_crf signal.
    """

    crf_status = models.CharField(
        verbose_name="CRF status",
        max_length=25,
        choices=DOCUMENT_STATUS,
        default=get_crf_status_default,
        help_text=_("If some data is still pending, flag this CRF as incomplete"),
    )

    crf_status_comments = models.TextField(
        verbose_name=_("Any comments related to status of this CRF"),
        blank=True,
        default="",
        help_text=_("for example, why some data is still pending"),
    )

    def update_crf_status_for_instance(self):
        return update_crf_status_for_instance(self)

    class Meta:
        abstract = True



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from dateutil.relativedelta import relativedelta
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist

from edc_utils import formatted_date, formatted_datetime, to_local
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.utils import get_lower_datetime

from ..constants import (
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    NEW_APPT,
    UNSCHEDULED_APPT,
)
from ..exceptions import (
    AppointmentInProgressError,
    AppointmentPermissionsRequired,
    AppointmentWindowError,
    CreateAppointmentError,
    InvalidParentAppointmentMissingVisitError,
    InvalidParentAppointmentStatusError,
    InvalidVisitCodeSequencesError,
    UnscheduledAppointmentError,
    UnscheduledAppointmentNotAllowed,
)
from .appointment_creator import AppointmentCreator

if TYPE_CHECKING:
    from edc_facility import Facility
    from edc_visit_schedule.visit import Visit

    from ..models import Appointment

__all__ = ["UnscheduledAppointmentCreator"]


class UnscheduledAppointmentCreator:
    """Attempts to create a new unscheduled appointment where the
    visit code sequence == the given `visit_code_sequence` or raises.
    """

    appointment_creator_cls = AppointmentCreator

    def __init__(
        self,
        *,
        subject_identifier: str,
        visit_schedule_name: str,
        schedule_name: str,
        visit_code: str,
        suggested_visit_code_sequence: int | None = None,
        suggested_appt_datetime: datetime | None = None,
        facility: Facility | None = None,
        request: Any | None = None,
    ):
        self._parent_appointment = None
        self._calling_appointment = None
        self._suggested_appt_datetime = None

        self.appointment = None

        self.subject_identifier = subject_identifier
        self.visit_schedule_name = visit_schedule_name
        self.schedule_name = schedule_name

        self.visit_code = visit_code
        self.visit_code_sequence = suggested_visit_code_sequence
        if self.visit_code_sequence is None:
            raise InvalidVisitCodeSequencesError("visit code sequence cannot be None")
        if self.visit_code_sequence < 1:
            raise InvalidVisitCodeSequencesError(
                "suggested visit code sequence cannot be less than 1"
            )
        self.facility = facility
        self.visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name)
        self.schedule = self.visit_schedule.schedules.get(schedule_name)
        self.appointment_model_cls = self.schedule.appointment_model_cls
        self.visit = self.visit_schedule.schedules.get(self.schedule_name).visits.get(
            self.visit_code
        )
        self.suggested_appt_datetime = suggested_appt_datetime
        self.has_perm_or_raise(request)
        self.create_or_raise()

    @property
    def visit(self) -> Visit:
        return self._visit

    @visit.setter
    def visit(self, value: Visit | None):
        if not value:
            raise UnscheduledAppointmentError(
                "Invalid visit. Got None using {"
                f"visit_schedule_name='{self.visit_schedule_name}',"
                f"schedule_name='{self.schedule_name}',"
                f"visit_code='{self.visit_code}'" + "}"
            )
        if not value.allow_unscheduled:
            raise UnscheduledAppointmentNotAllowed(
                f"Not allowed. Visit {self.visit_code} is not configured for "
                "unscheduled appointments."
            )
        self._visit = value

    def create_or_raise(self) -> None:
        """Create the unscheduled appointment.

        Do not allow new unscheduled if any appointments are IN_PROGRESS

        Note: `timepoint` and `timepoint_datetime` are not incremented
            for unscheduled appointments. These values are carried
            over from the parent appointment values.

        """
        try:
            obj = self.appointment_model_cls.objects.get(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                appt_status=IN_PROGRESS_APPT,
            )
        except MultipleObjectsReturned as e:
            raise UnscheduledAppointmentError(e) from e
        except ObjectDoesNotExist:
            pass
        else:
            raise AppointmentInProgressError(
                f"Not allowed. Appointment {obj.visit_code}."
                f"{obj.visit_code_sequence} is in progress."
            )
        try:
            appointment_creator = self.appointment_creator_cls(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit=self.visit,
                suggested_datetime=self.suggested_appt_datetime,
                timepoint=self.parent_appointment.timepoint,
                timepoint_datetime=self.parent_appointment.timepoint_datetime,
                visit_code_sequence=self.parent_appointment.next_visit_code_sequence,
                facility=self.facility,
                appt_status=NEW_APPT,
                appt_reason=UNSCHEDULED_APPT,
                ignore_window_period=self.ignore_window_period,
            )
        except CreateAppointmentError:
            raise
        except AppointmentWindowError as e:
            msg = str(e).replace("Perhaps catch this in the form", "")
            raise UnscheduledAppointmentError(
                f"Unable to create unscheduled appointment. {msg}"
            ) from e
        self.appointment = appointment_creator.appointment

    def has_perm_or_raise(self, request) -> None:
        if request and not request.user.has_perm(
            f"{self.appointment_model_cls._meta.app_label}."
            f"add_{self.appointment_model_cls._meta.model_name}"
        ):
            raise AppointmentPermissionsRequired(
                "You do not have permission to create an appointment"
            )

    @property
    def suggested_appt_datetime(self):
        return self._suggested_appt_datetime

    def after_calling_appt_or_raise(self, suggested_dte: datetime):
        """Raises if on same day or before otherwise returns True"""
        suggested_dt = to_local(suggested_dte).date()
        calling_dt = to_local(self.calling_appointment.appt_datetime).date()
        if suggested_dt <= calling_dt:
            suggested = formatted_date(suggested_dt)
            calling = formatted_date(calling_dt)
            raise CreateAppointmentError(
                "Suggested appointment date must be after the calling appointment date. "
                f"Got {suggested} not after {calling}."
            )
        return True

    @suggested_appt_datetime.setter
    def suggested_appt_datetime(self, suggested_dte: datetime | None):
        if suggested_dte and self.after_calling_appt_or_raise(suggested_dte):
            self._suggested_appt_datetime = suggested_dte
        else:
            self._suggested_appt_datetime = (
                self.calling_appointment.appt_datetime + relativedelta(days=1)
            )
        if self.parent_appointment.next and self.suggested_appt_datetime >= get_lower_datetime(
            self.parent_appointment.next
        ):
            dt = formatted_datetime(self.suggested_appt_datetime)
            next_dt = formatted_datetime(get_lower_datetime(self.parent_appointment.next))
            raise UnscheduledAppointmentError(
                "Appointment date exceeds window period. Next appointment is "
                f"{self.parent_appointment.next.visit_code} and lower window starts "
                f"on {next_dt}. Got {dt}. Perhaps catch this in the form."
            )

    @property
    def calling_appointment(self) -> Appointment:
        """Returns the appointment which precedes the to-be-created
        unscheduled appt.

        It must exist.

        Note: subtracting 1 from `visit_code_sequence` implies the
        visit_code_sequence cannot be less than 1.

        if visit_code_sequence = 1, calling_appointment and parent_appointment
        will be the same instance.
        """
        if not self._calling_appointment:
            opts = dict(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence=self.visit_code_sequence - 1,
                timepoint=self.parent_appointment.timepoint,
            )
            self._calling_appointment = self.appointment_model_cls.objects.get(**opts)
        return self._calling_appointment

    @property
    def parent_appointment(self) -> Appointment:
        """Returns the parent or 'scheduled' appointment
        (visit_code_sequence=0).
        """
        if not self._parent_appointment:
            options = dict(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence=0,
            )
            self._parent_appointment = self.appointment_model_cls.objects.get(**options)
            if not self._parent_appointment.related_visit:
                raise InvalidParentAppointmentMissingVisitError(
                    "Unable to create unscheduled appointment. An unscheduled "
                    "appointment cannot be created if the parent appointment's "
                    "visit form is not submitted. "
                    f"Got appointment '{self.visit_code}.0'."
                )
            if self._parent_appointment.appt_status not in [
                COMPLETE_APPT,
                INCOMPLETE_APPT,
            ]:
                raise InvalidParentAppointmentStatusError(
                    "Unable to create unscheduled appointment. An unscheduled "
                    "appointment cannot be created if the parent appointment "
                    "is 'new' or 'in progress'. Got appointment "
                    f"'{self.visit_code}' is "
                    f"{self._parent_appointment.get_appt_status_display().lower()}."
                )

        return self._parent_appointment

    @property
    def ignore_window_period(self: Any) -> bool:
        value = False
        if (
            self.calling_appointment
            and self.calling_appointment.next
            and self.calling_appointment.next.appt_status in [INCOMPLETE_APPT, COMPLETE_APPT]
            and self.suggested_appt_datetime < self.calling_appointment.next.appt_datetime
        ):
            value = True
        return value



from typing import Any

from clinicedc_constants import FASTING, YES
from django import forms

from edc_glucose.utils import validate_glucose_as_millimoles_per_liter


class BloodResultsFbgFormValidatorMixin:
    @property
    def reportables_evaluator_options(self: Any):
        if not self.cleaned_data.get("fasting"):
            raise forms.ValidationError({"fasting": "This field is required."})
        fasting = bool(
            self.cleaned_data.get("fasting") == FASTING
            or self.cleaned_data.get("fasting") == YES
        )
        return dict(fasting=fasting)

    def evaluate_value(self, prefix: str = None):
        validate_glucose_as_millimoles_per_liter(prefix=prefix, cleaned_data=self.cleaned_data)



# Generated by Django 5.1.2 on 2024-11-17 14:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0032_historicalstockrequest_task_id_stockrequest_task_id"),
    ]

    operations = [
        migrations.AddField(
            model_name="container",
            name="display_name",
            field=models.CharField(blank=True, max_length=100, null=True, unique=True),
        ),
        migrations.AddField(
            model_name="historicalcontainer",
            name="display_name",
            field=models.CharField(blank=True, db_index=True, max_length=100, null=True),
        ),
    ]



from __future__ import annotations

from decimal import Decimal

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import PROTECT
from django.utils import timezone
from edc_model.models import BaseUuidModel, HistoricalRecords
from sequences import get_next_value

from ...choices import STOCK_STATUS
from ...constants import ALLOCATED, AVAILABLE, ZERO_ITEM
from ...exceptions import AllocationError, AssignmentError, StockError
from ...utils import get_random_code
from .allocation import Allocation
from .container import Container
from .location import Location
from .lot import Lot
from .managers import StockManager
from .product import Product
from .receive_item import ReceiveItem
from .repack_request import RepackRequest


class Stock(BaseUuidModel):
    stock_identifier = models.CharField(
        verbose_name="Internal stock identifier",
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    code = models.CharField(
        verbose_name="Stock code",
        max_length=15,
        unique=True,
        null=True,
        blank=True,
        help_text="A unique alphanumeric code",
    )

    stock_datetime = models.DateTimeField(
        default=timezone.now, help_text="date stock record created"
    )

    receive_item = models.ForeignKey(
        ReceiveItem, on_delete=models.PROTECT, null=True, blank=False
    )

    repack_request = models.ForeignKey(
        RepackRequest, on_delete=models.PROTECT, null=True, blank=True
    )

    from_stock = models.ForeignKey(
        "edc_pharmacy.stock",
        related_name="source_stock",
        on_delete=models.PROTECT,
        null=True,
    )

    confirmed = models.BooleanField(
        default=False,
        help_text=(
            "True if stock was labelled and confirmed; "
            "False if stock was received/repacked but never confirmed."
        ),
    )
    confirmed_datetime = models.DateTimeField(null=True, blank=True)

    confirmed_by = models.CharField(max_length=150, default="", blank=True)

    allocation = models.OneToOneField(
        Allocation,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        help_text="Subject allocation",
    )

    product = models.ForeignKey(Product, on_delete=models.PROTECT)

    lot = models.ForeignKey(
        Lot, verbose_name="Batch", on_delete=models.PROTECT, null=True, blank=False
    )

    container = models.ForeignKey(Container, on_delete=models.PROTECT, null=True, blank=False)

    location = models.ForeignKey(Location, on_delete=PROTECT, null=True, blank=False)

    qty = models.DecimalField(
        null=True,
        blank=False,
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
        help_text="Difference of qty_in and qty_out",
    )

    qty_in = models.DecimalField(
        null=True,
        blank=False,
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        help_text="Container qty, e.g. 1 bucket, 1 bottle, etc",
    )

    qty_out = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        help_text="Container qty, e.g. 1 bucket, 1 bottle, etc",
    )

    unit_qty_in = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
        validators=[MinValueValidator(0)],
        help_text="Number of units in this container, e.g. 128 tablets. See post-save signal",
    )

    unit_qty_out = models.DecimalField(
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
        validators=[MinValueValidator(0)],
        help_text="Number of units from this container, e.g. 128 tablets",
    )

    status = models.CharField(max_length=25, choices=STOCK_STATUS, default=AVAILABLE)

    description = models.CharField(max_length=100, default="", blank=True)

    in_transit = models.BooleanField(default=False, help_text="See stocktransferitem.")

    confirmed_at_location = models.BooleanField(
        default=False, help_text="See confirmeatlocationitem."
    )

    stored_at_location = models.BooleanField(default=False, help_text="See storagebinitem.")

    dispensed = models.BooleanField(default=False)

    destroyed = models.BooleanField(default=False)

    subject_identifier = models.CharField(
        max_length=50, default="", blank=True, editable=False
    )

    objects = StockManager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.code}: {self.description}"

    def save(self, *args, **kwargs):
        if not self.stock_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.stock_identifier = f"{next_id:010d}"
            self.code = get_random_code(self.__class__, 6, 10000)
            self.product = self.get_receive_item().order_item.product
        if not self.description:
            self.description = f"{self.product.name} - {self.container.name}"
        self.verify_assignment_or_raise()
        self.verify_assignment_or_raise(self.from_stock)
        self.update_status()

        # in_transit
        if "in_transit" not in kwargs.get("update_fields", []):
            original_instance = Stock.objects.get(pk=self.pk)
            if self.in_transit != original_instance.in_transit:
                raise StockError(
                    "Invalid attempt to change field. The value of field `in_transit` "
                    "is only set in the post-save/delete signals of model StockTransferItem."
                )

        # received / confirmed at location

        # stored_at_location
        if "stored_at_location" not in kwargs.get("update_fields", []):
            original_instance = Stock.objects.get(pk=self.pk)
            if self.stored_at_location != original_instance.stored_at_location:
                raise StockError(
                    "Invalid attempt to change field. The value of field `stored_at_location` "
                    "is only set in the post-save/delete signals of model StorageBinItem."
                )

        # dispensed
        if "dispensed" not in kwargs.get("update_fields", []):
            original_instance = Stock.objects.get(pk=self.pk)
            if self.dispensed != original_instance.dispensed:
                raise StockError(
                    "Invalid attempt to change field. The value of field `dispensed` "
                    "is only set in the post-save/delete signals of model DispenseItem."
                )

        # destroyed

        # do this in the post-save signal?
        # self.unit_qty_in = Decimal(self.qty_in) * Decimal(self.container.qty)
        super().save(*args, **kwargs)

    def update_transferred(self) -> bool:
        return (
            self.allocation
            and self.allocation.stock_request_item.stock_request.location == self.location
            and self.container.may_request_as
        )

    def verify_assignment_or_raise(
        self, stock: models.ForeignKey[Stock] | None = None
    ) -> None:
        """Verify that the LOT and PRODUCT assignments match."""
        if not stock:
            stock = self
        if stock.product.assignment != stock.lot.assignment:
            raise AssignmentError("Lot number assignment does not match product assignment!")
        if self.allocation and self.allocation.assignment != stock.lot.assignment:
            raise AllocationError(
                f"Allocation assignment does not match lot assignment! Got {self.code}."
            )

    def update_status(self):
        if self.allocation:
            self.status = ALLOCATED
        elif self.qty_out == self.qty_in:
            self.status = ZERO_ITEM
        else:
            self.status = AVAILABLE

    def get_receive_item(self) -> ReceiveItem:
        """Recursively fetch the original receive item."""
        obj: Stock = self
        receive_item = self.receive_item
        while not receive_item:
            obj = obj.from_stock
            receive_item = obj.receive_item
        return receive_item

    @property
    def unit_qty(self):
        """Unit qty on hand"""
        return self.unit_qty_in - self.unit_qty_out

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock"
        verbose_name_plural = "Stock"



from .data_request import DataRequest
from .data_request_history import DataRequestHistory
from .edc_permissions import EdcPermissions
from .permission_dummies import ExportData, ImportData
from .signals import (
    export_transaction_history_on_post_save,
    export_transaction_history_on_pre_delete,
)

__all__ = [
    "DataRequest",
    "DataRequestHistory",
    "EdcPermissions",
    "ExportData",
    "ImportData",
    "export_transaction_history_on_post_save",
    "export_transaction_history_on_pre_delete",
]



from django.db import models
from django.db.models.deletion import PROTECT
from django.utils import timezone

from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)
from edc_action_item.models import ActionModelMixin
from edc_constants.choices import YES_NO
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import ReportStatusModelMixin
from edc_model.validators import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_sites.model_mixins import SiteModelMixin

from ...constants import DEATH_REPORT_TMG_ACTION, DEATH_REPORT_TMG_SECOND_ACTION
from ...models import CauseOfDeath
from ...utils import get_adverse_event_app_label


class DeathReportTmgManager(ActionIdentifierModelManager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(action_item__action_type__name=DEATH_REPORT_TMG_ACTION)


class DeathReportTmgSiteManager(ActionIdentifierSiteManager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(action_item__action_type__name=DEATH_REPORT_TMG_ACTION)


class DeathReportTmgSecondManager(ActionIdentifierModelManager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(action_item__action_type__name=DEATH_REPORT_TMG_SECOND_ACTION)


class DeathReportTmgSecondSiteManager(ActionIdentifierSiteManager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(action_item__action_type__name=DEATH_REPORT_TMG_SECOND_ACTION)


class DeathReportTmgFieldsModelMixin(models.Model):
    death_report = models.ForeignKey(
        f"{get_adverse_event_app_label()}.deathreport", on_delete=PROTECT
    )

    report_datetime = models.DateTimeField(
        verbose_name="Report Date",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=timezone.now,
    )

    cause_of_death = models.ForeignKey(
        CauseOfDeath,
        on_delete=PROTECT,
        verbose_name="Main cause of death",
        help_text=(
            "Main cause of death in the opinion of the local study doctor and local PI"
        ),
        null=True,
        default="",
    )

    cause_of_death_other = models.CharField(
        verbose_name='If "Other" above, please specify',
        max_length=100,
        blank=True,
        default="",
    )

    cause_of_death_agreed = models.CharField(
        verbose_name="Is the cause of death agreed between study doctor and TMG member?",
        max_length=15,
        choices=YES_NO,
        default="",
        help_text="If No, explain in the narrative below",
    )

    narrative = models.TextField(verbose_name="Narrative", blank=True, default="")

    class Meta:
        abstract = True


class DeathReportTmgMethodsModelMixin(models.Model):
    class Meta:
        abstract = True

    def __str__(self):
        return str(self.death_report)

    def save(self, *args, **kwargs):
        self.subject_identifier = self.death_report.subject_identifier
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.action_identifier,)

    natural_key.dependencies = ["edc_adverse_event.causeofdeath"]


class DeathReportTmgModelMixin(
    SiteModelMixin,
    ActionModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    DeathReportTmgFieldsModelMixin,
    DeathReportTmgMethodsModelMixin,
    ReportStatusModelMixin,
    models.Model,
):
    action_name = DEATH_REPORT_TMG_ACTION

    objects = DeathReportTmgManager()

    on_site = DeathReportTmgSiteManager()

    class Meta(
        SiteModelMixin.Meta,
        NonUniqueSubjectIdentifierFieldMixin.Meta,
        ActionModelMixin.Meta,
    ):
        abstract = True
        verbose_name = "Death Report TMG (1st)"
        verbose_name_plural = "Death Report TMG (1st)"
        indexes = (
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
            *ActionModelMixin.Meta.indexes,
            models.Index(fields=["subject_identifier", "action_identifier", "site", "id"]),
        )



# from __future__ import annotations
#
# from typing import TYPE_CHECKING
#
# from .subquery import Subquery
#
# if TYPE_CHECKING:
#     from .qa_case import QaCase
#
#
# def subquery_from_dict(
#     cases: list[dict[str:str, str:str, str:str] | QaCase],
#     as_list: bool | None = False,
# ) -> str | list:
#     """Returns an SQL select statement as a union of the select
#     statements of each case.
#
#     args:
#      cases = [{
#          "label_lower": "my_app.hivhistory",
#          "dbtable": "my_app_hivhistory",
#          "field": "hiv_init_date",
#          "label": "missing HIV initiation date",
#          "list_tables": [(list_field, list_dbtable, alias), ...],
#          }, ...]
#
#          Note: `list_field` is the CRF id field, for example:
#             left join <list_dbtable> as <alias> on crf.<list_field>=<alias>.id
#     """
#     subqueries = []
#     for case in cases:
#         try:
#             subquery = case.sql
#         except AttributeError:
#             subquery = Subquery(**case).sql
#         subqueries.append(subquery)
#     if as_list:
#         return subqueries
#     return " UNION ".join(subqueries)



from __future__ import annotations

import contextlib
from datetime import date

from django import forms
from django.utils.translation import gettext as _

from edc_utils.date import to_local

from .base_form_validator import BaseFormValidator


def convert_any_to_date(date1, date2) -> tuple[date, date]:
    with contextlib.suppress(AttributeError):
        date1 = to_local(date1).date()
    with contextlib.suppress(AttributeError):
        date2 = to_local(date2).date()
    return date1, date2


class DateRangeFieldValidator(BaseFormValidator):
    def date_not_before(
        self,
        date_field1: str,
        date_field2: str,
        msg=None,
        convert_to_date: bool | None = None,
        message_on_field=None,
    ) -> None:
        """Asserts date_field2 is not before date_field1"""
        date1 = self.cleaned_data.get(date_field1)
        date2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            date1, date2 = convert_any_to_date(date1, date2)
        if date1 and date2 and (date2 < date1):
            msg = msg or _("Invalid. Cannot be before %(field)s.") % {"field": date_field1}
            raise forms.ValidationError({message_on_field or date_field2: msg})

    def date_not_after(
        self,
        date_field1: str,
        date_field2: str,
        msg: str | None = None,
        convert_to_date: bool | None = None,
    ) -> None:
        """Asserts date_field2 is not after date_field1"""
        date1 = self.cleaned_data.get(date_field1)
        date2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            date1, date2 = convert_any_to_date(date1, date2)
        if date1 and date2 and (date2 > date1):
            msg = msg or _("Invalid. Cannot be after %(field)s.") % {"field": date_field1}
            raise forms.ValidationError({date_field2: msg})

    def date_equal(
        self,
        date_field1: str,
        date_field2: str,
        msg=None,
        message_on_field=None,
        convert_to_date: bool | None = None,
    ) -> None:
        """Asserts date2 and date1 are equal"""
        date1 = self.cleaned_data.get(date_field1)
        date2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            date1, date2 = convert_any_to_date(date1, date2)
        if date1 and date2 and (date1 != date2):
            msg = msg or _("Invalid. Expected %(field2)s to be the same as %(field1)s.") % {
                "field1": date_field1,
                "field2": date_field2,
            }
            raise forms.ValidationError({message_on_field or date_field2: msg})

    def date_not_equal(
        self,
        date_field1: str,
        date_field2: str,
        msg=None,
        message_on_field=None,
        convert_to_date: bool | None = None,
    ) -> None:
        """Asserts date2 and date1 are not equal"""
        dte1 = self.cleaned_data.get(date_field1)
        dte2 = self.cleaned_data.get(date_field2)
        if convert_to_date:
            dte1, dte2 = convert_any_to_date(dte1, dte2)
        if dte1 and dte2 and (dte1 == dte2):
            msg = msg or _("Invalid. Expected %(field2)s to be different from %(field1)s.") % {
                "field1": date_field1,
                "field2": date_field2,
            }
            raise forms.ValidationError({message_on_field or date_field2: msg})

    def datetime_not_before(
        self, datetime_field1: str, datetime_field2: str, msg=None
    ) -> None:
        """Asserts datetime_field2 is not before datetime_field1"""
        dte1 = self.cleaned_data.get(datetime_field1)
        dte2 = self.cleaned_data.get(datetime_field2)
        if dte1 and dte2 and (dte2 < dte1):
            msg = msg or _("Invalid. Cannot be before %(field)s") % {"field": datetime_field1}
            raise forms.ValidationError({datetime_field2: msg})

    def datetime_not_after(self, datetime_field1: str, datetime_field2: str, msg=None) -> None:
        """Asserts datetime_field2 is not after datetime_field1"""
        dte1 = self.cleaned_data.get(datetime_field1)
        dte2 = self.cleaned_data.get(datetime_field2)
        if dte1 and dte2 and (dte2 > dte1):
            msg = msg or _("Invalid. Cannot be after %(field)s") % {"field": datetime_field1}
            raise forms.ValidationError({datetime_field2: msg})

    def datetime_equal(self, datetime_field1: str, datetime_field2: str, msg=None) -> None:
        """Asserts datetime_field2 is not equal to datetime_field1"""
        dte1 = self.cleaned_data.get(datetime_field1)
        dte2 = self.cleaned_data.get(datetime_field2)
        if dte1 and dte2 and (dte1 == dte2):
            msg = msg or _("Invalid. Cannot be same as %(field)s") % {"field": datetime_field1}
            raise forms.ValidationError({datetime_field2: msg})



from .allocation_admin import AllocationAdmin
from .allocation_proxy_admin import AllocationProxyAdmin
from .confirmation_admin import ConfirmationAdmin
from .confirmation_at_location_admin import ConfirmationAtLocationAdmin
from .confirmation_at_location_item_admin import ConfirmationAtLocationItemAdmin
from .container_admin import ContainerAdmin
from .container_type_admin import ContainerTypeAdmin
from .dispense_admin import DispenseAdmin
from .dispense_item_admin import DispenseItemAdmin
from .history import (
    ConfirmationAtLocationHistoryAdmin,
    ConfirmationAtLocationItemHistoryAdmin,
    StorageBinItemHistoryAdmin,
)
from .location_admin import LocationAdmin
from .lot_admin import LotAdmin
from .order_admin import OrderAdmin
from .order_item_admin import OrderItemAdmin
from .product_admin import ProductAdmin
from .receive_admin import ReceiveAdmin
from .receive_item_admin import ReceiveItemAdmin
from .repack_request_admin import RequestRepackAdmin
from .stock_adjustment_admin import StockAdjustmentAdmin
from .stock_admin import StockAdmin
from .stock_proxy_admin import StockProxyAdmin
from .stock_request_admin import StockRequestAdmin
from .stock_request_item_admin import StockRequestItemAdmin
from .stock_transfer_admin import StockTransferAdmin
from .stock_transfer_item_admin import StockTransferItemAdmin
from .storage_bin_admin import StorageBinAdmin
from .storage_bin_item_admin import StorageBinItemAdmin
from .supplier_admin import SupplierAdmin



# Generated by Django 3.0.6 on 2020-07-29 20:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_crf", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="crfstatus",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "CRF Status",
                "verbose_name_plural": "CRF Status",
            },
        ),
    ]



from django.urls import path
from django.views.generic import RedirectView

app_name = "edc_protocol_incident"

urlpatterns = [
    path("", RedirectView.as_view(url="/edc_protocol_incident/admin/"), name="home_url"),
]



import sys

from django.core.management.color import color_style

style = color_style()


def post_migrate_update_sites(sender=None, **kwargs):
    from .site import sites as site_sites  # noqa: PLC0415
    from .utils import add_or_update_django_sites  # noqa: PLC0415

    sys.stdout.write(style.MIGRATE_HEADING("Updating sites:\n"))

    for country in site_sites.countries:
        sys.stdout.write(style.MIGRATE_HEADING(f" (*) sites for {country} ...\n"))
        add_or_update_django_sites(verbose=True)
    sys.stdout.write("Done.\n")
    sys.stdout.flush()



from django.db import models

from edc_action_item.models import ActionItem as BaseActionItem

from ..action_items import DATA_QUERY_ACTION


class Manager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(action_type__name__in=[DATA_QUERY_ACTION])


class DataManagerActionItem(BaseActionItem):
    objects = Manager()

    class Meta:
        proxy = True
        default_permissions = ("view",)



from __future__ import annotations

import sys
from typing import Any

from django.apps import apps as django_apps
from django.contrib.admin.sites import all_sites
from django.db.models import Count
from tqdm import tqdm

from edc_appointment.models import Appointment
from edc_metadata.models import CrfMetadata, RequisitionMetadata
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.utils import get_related_visit_model_cls

from .metadata import CrfMetadataGetter, RequisitionMetadataGetter
from .metadata_rules import site_metadata_rules


class MetadataRefresher:
    """A class to be `run` when metadata gets out-of-date

    This may happen when there are changes to the visit schedule,
    metadata rules or manual changes to data.
    """

    def __init__(self, verbose: bool | None = None):
        self._source_models = []
        self._admin_models = []
        self.verbose = verbose

    def run(self) -> None:
        self._message("Updating metadata ...     \n")
        self.create_or_update_metadata_for_all()

        # note: only need to run metadata rules on the related
        # visit model
        self._message("Running metadata rules ...\n")
        total = get_related_visit_model_cls().objects.all().count()
        self.run_metadata_rules(get_related_visit_model_cls(), total)
        self._message("Done.\n")

    @property
    def source_models(self) -> list[str]:
        if not self._source_models:
            self._source_models = []
            for rule_groups_list in site_metadata_rules.rule_groups.values():
                for rule_groups in rule_groups_list:
                    if (
                        rule_groups._meta.source_model
                        != get_related_visit_model_cls()._meta.label_lower
                    ):
                        self._source_models.append(rule_groups._meta.source_model)
            self._source_models = list(set(self._source_models))
            self._source_models.sort()
            self._source_models.insert(0, get_related_visit_model_cls()._meta.label_lower)
            self._message(f"  Found source models: {', '.join(self.source_models)}.\n")
        return self._source_models

    @staticmethod
    def run_metadata_rules(source_model_cls: Any, total: int) -> None:
        """Updates rules for all instances of this source model"""
        for instance in tqdm(source_model_cls.objects.all(), total=total):
            if django_apps.get_app_config("edc_metadata").metadata_rules_enabled:
                instance.run_metadata_rules()

    def create_or_update_metadata(self, source_model_cls: Any, total: int) -> None:
        """Creates or updates CRF/Requisition metadata for all instances
        of this source model.
        """
        for instance in tqdm(source_model_cls.objects.all(), total=total):
            try:
                instance.metadata_create()
            except AttributeError:
                try:
                    instance.metadata_update()
                except AttributeError as e:
                    self._message(f"      skipping (got {e})     \n")
                    break

    def _message(self, msg: str) -> None:
        if self.verbose:
            sys.stdout.write(msg)

    @property
    def crf_metadata_models(self) -> list[CrfMetadata]:
        return [
            obj.get("model")
            for obj in CrfMetadata.objects.values("model")
            .order_by("model")
            .annotate(count=Count("model"))
        ]

    @property
    def requisition_metadata_models(self) -> list[RequisitionMetadata]:
        return [
            obj.get("model")
            for obj in RequisitionMetadata.objects.values("model")
            .order_by("model")
            .annotate(count=Count("model"))
        ]

    @property
    def admin_models(self) -> list:
        if not self._admin_models:
            for admin_site in all_sites:
                self._admin_models.extend(
                    [cls._meta.label_lower for cls in admin_site._registry]
                )
        return self._admin_models

    def verifying_crf_metadata_with_visit_schedule_and_admin(self) -> None:
        self._message("- Verifying CrfMetadata models with visit schedule and admin.\n")
        for model in self.crf_metadata_models:
            if (
                model not in site_visit_schedules.all_post_consent_models
                and model not in self.admin_models
            ):
                count = CrfMetadata.objects.filter(model=model).delete()
                self._message(f"   * deleted {count} metadata records for model {model}.\n")
        self._message("    Done.\n")

    def verifying_requisition_metadata_with_visit_schedule_and_admin(self) -> None:
        self._message(
            "- Verifying RequisitionMetadata models with visit schedule and admin.\n"
        )
        for model in self.requisition_metadata_models:
            if (
                model not in site_visit_schedules.all_post_consent_models
                and model not in self.admin_models
            ):
                count = RequisitionMetadata.objects.filter(model=model).delete()
                self._message(f"   * deleted {count} metadata records for model {model}.\n")
        self._message("    Done.\n")

    def create_or_update_metadata_for_all(self) -> None:
        self.verifying_crf_metadata_with_visit_schedule_and_admin()
        self.verifying_requisition_metadata_with_visit_schedule_and_admin()
        # TODO: perhaps always delete before running
        self._message("- Updating metadata for all post consent models...     \n")
        models = dict(sorted(site_visit_schedules.all_post_consent_models.items()))
        model_count = len(list(models))
        related_visits = get_related_visit_model_cls().objects.all()
        total = related_visits.count()
        self._message(
            f"   - {model_count} post-consent models found for {total} visits ... \n"
        )
        for related_visit in tqdm(related_visits, total=total):
            related_visit.metadata_create()
        self._message("    Done.\n")

    def validate_metadata_for_all(self):
        self._message("- Validating all metadata ...     \n")
        appointments = Appointment.objects.all().order_by("site_id")
        total = appointments.count()
        for appointment in tqdm(appointments, total=total):
            CrfMetadataGetter(appointment=appointment)
            RequisitionMetadataGetter(appointment=appointment)



from django.db import models
from django.db.models import Index

from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel
from edc_visit_schedule.model_mixins import (
    VisitCodeFieldsModelMixin,
    VisitScheduleFieldsModelMixin,
)


class CrfStatus(
    NonUniqueSubjectIdentifierFieldMixin,
    VisitScheduleFieldsModelMixin,
    VisitCodeFieldsModelMixin,
    BaseUuidModel,
):
    label_lower = models.CharField(max_length=150, null=True)

    class Meta(BaseUuidModel.Meta, NonUniqueSubjectIdentifierFieldMixin.Meta):
        verbose_name = "CRF Status"
        verbose_name_plural = "CRF Status"
        indexes = (
            *BaseUuidModel.Meta.indexes,
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
            Index(
                fields=[
                    "schedule_name",
                    "visit_schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                ]
            ),
        )



from clinicedc_constants import CLOSED, NEW, NORMAL, OPEN, RESOLVED
from django.db import models
from django.db.models import PROTECT
from django.utils import timezone

from ..choices import DM_STATUS, QUERY_PRIORITY, RESPONSE_STATUS
from ..constants import CLOSED_WITH_ACTION
from .user import DataManagerUser, QueryUser


class DataQueryModelMixin(models.Model):
    report_datetime = models.DateTimeField(verbose_name="Query date", default=timezone.now)

    title = models.CharField(max_length=150, default="", blank=False)

    sender = models.ForeignKey(
        DataManagerUser,
        related_name="+",
        on_delete=PROTECT,
        verbose_name="Query raised by",
        help_text="Please select a name from the list",
    )

    recipients = models.ManyToManyField(
        QueryUser,
        related_name="+",
        verbose_name="Sent to",
        help_text=(
            "Select any additional recipients. Users in the `Site Data Manager` "
            "group are automatically included."
        ),
        blank=True,
    )

    query_priority = models.CharField(
        verbose_name="Priority", max_length=25, choices=QUERY_PRIORITY, default=NORMAL
    )

    query_text = models.TextField(help_text="Describe the query in detail.")

    site_resolved_datetime = models.DateTimeField(
        verbose_name="Site resolved on", null=True, blank=True
    )

    site_response_text = models.TextField(default="", blank=True)

    site_response_status = models.CharField(
        verbose_name="Site status", max_length=25, choices=RESPONSE_STATUS, default=NEW
    )

    status = models.CharField(
        verbose_name="DM status", max_length=25, choices=DM_STATUS, default=OPEN
    )

    dm_user = models.ForeignKey(
        DataManagerUser,
        verbose_name="DM resolved by",
        related_name="+",
        on_delete=PROTECT,
        null=True,
        blank=True,
        help_text="Please select a name from the list",
    )

    resolved_datetime = models.DateTimeField(
        verbose_name="DM resolved on", null=True, blank=True
    )

    auto_resolved = models.BooleanField(default=False)

    plan_of_action = models.TextField(
        default="", blank=True, help_text="If required, provide a plan of action"
    )

    locked = models.BooleanField(
        default=False,
        help_text="If locked, this query will NEVER be reopened.",
    )

    locked_reason = models.TextField(
        verbose_name="Reason query locked",
        default="",
        blank=True,
        help_text="If required, the reason the query cannot be resolved.",
    )

    @property
    def dm_resolved(self):
        return self.status in [CLOSED, CLOSED_WITH_ACTION]

    @property
    def site_resolved(self):
        return self.site_response_status == RESOLVED

    class Meta:
        abstract = True
